
{-# LANGUAGE OverloadedStrings #-} 
-- {-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE TemplateHaskell #-}
import Data.Aeson
import Data.Aeson.Lens
--import Text.ParserCombinators.Parsec
import qualified Data.ByteString.Char8 as BC8
import qualified Data.ByteString.Lazy as B
import qualified Data.Binary.Builder as BB
import Network.Socket
import Network.Http.Client
--import Network.Http.Connection (Connection)
import System.IO.Streams (InputStream, OutputStream, stdout)
import qualified System.IO.Streams as Streams
--import GHC.Generics
import Control.Lens hiding ((.=))
import qualified Control.Lens as L --hiding (element)
import Data.Maybe
import Control.Monad
import Control.Monad.State hiding (put, get)
import qualified Control.Monad.State as St
--import System.IO
--import GHC.IO.Handle.FD
import Data.Time.Clock.POSIX
--import Data.Ratio (numerator)
import Control.Concurrent (threadDelay)
import qualified Data.Text as T
import System.Environment (getArgs)

getKodiActivePlayers = BC8.readFile "jsons/getActivePlayers.json"

getKodiPlayerPositionTime :: Int -> IO BC8.ByteString
getKodiPlayerPositionTime playerid = do 
   f<- BC8.readFile "jsons/getPlayerPositionTime.json"
   return $ f & key "params" . key "playerid" . _JSON .~ playerid
   
getKodiPlayerPositionTimeP :: Int -> BC8.ByteString -> IO BC8.ByteString
getKodiPlayerPositionTimeP playerid f = do 
   --f<- BC8.readFile "jsons/getPlayerPositionTime.json"
   return $ f & key "params" . key "playerid" . _JSON .~ playerid
   
parseForKodiPlayerid :: BC8.ByteString -> Maybe Int
parseForKodiPlayerid s = do
   s ^? key "result" . nth 0 . key "playerid" . _Integer >>= return.fromInteger

parseForKodiPlayerPositionTime :: BC8.ByteString -> Maybe (Int, Int, Int, Int)
parseForKodiPlayerPositionTime s = do
   time $ sequence [
       s ^? key "result" . key "time" . key "hours" . _Integer >>= return.fromInteger
      ,s ^? key "result" . key "time" . key "minutes" . _Integer >>= return.fromInteger
      ,s ^? key "result" . key "time" . key "seconds" . _Integer >>= return.fromInteger
      ,s ^? key "result" . key "time" . key "milliseconds" . _Integer >>= return.fromInteger
     ]
   where
   time (Just [_1,_2,_3,_4]) = Just (_1,_2,_3,_4)
   time _ = Nothing


parseForMpvTime_pos :: BC8.ByteString -> Maybe Float
parseForMpvTime_pos s = parseMpvResponceDataFloat s

parseMpvResponceDataFloat :: BC8.ByteString -> Maybe Float
parseMpvResponceDataFloat s = do
   time (
       s ^? key "error" . _String 
      ,s ^? key "data" . _Double
     )
   where
   time (Just "success", Just _2) = Just $ realToFrac _2
   time _ = Nothing

parseForAudioSpeedCorrection :: BC8.ByteString -> Maybe Float
parseForAudioSpeedCorrection s = parseMpvResponceDataFloat s
   
   
   
data A = A 
   { _response :: Maybe BC8.ByteString
    ,_playerid :: Maybe Int
    ,_adjThreshold :: Float
    ,_desync :: Float
    ,_pause :: Bool
    ,_nextWakeupTime :: POSIXTime
    ,_kodiPlayerPositionTime :: Maybe (Int, Int, Int, Int)
    ,_mpvTime_pos :: Maybe Float
    ,_mpvSpeed :: Maybe Float
    ,_mpvSocket :: FilePath
    ,_kodiSocketHostname :: Hostname
    ,_kodiSocketPort :: Port
    ,_kodiConnection :: Connection
    ,_mpvConnection  :: Socket
    ,_mpvCommError :: Maybe IOError
    ,_verbose :: Int
   }
L.makeLenses ''A



getTime_posMPV :: StateT A IO ()
getTime_posMPV = do 
   r<- St.get
   
   when (r^.verbose>2) $  liftIO $ putStrLn "getTime_posMPV"
   
   getPlayerPositionTimeMPV' <- liftIO $ readFile "jsons/getTime_posMPV.json"
   
   liftIO $ send (r^.mpvConnection) getPlayerPositionTimeMPV'
   (liftIO $ recv (r^.mpvConnection) 4096) 
      >>= (return . responseWrapper) 
      >>= (response L..= )
   
   r<- St.get
   when (r^.verbose>0) $ liftIO $ BC8.putStrLn $ fromMaybe "---" $ r^.response 
   
   L.assign mpvTime_pos $ maybe Nothing parseForMpvTime_pos (r^.response) 
   r<- St.get
   when (r^.verbose>2) $  liftIO $ print $ r^.mpvTime_pos
   
   where
   responseWrapper "" = Nothing
   responseWrapper x  = Just $ BC8.pack x
   

getAudioSpeedCorrectionMPV :: StateT A IO ()
getAudioSpeedCorrectionMPV = do 
   r<- St.get
   
   when (r^.verbose>2) $  liftIO $ putStrLn "getAudioSpeedCorrectionMPV"
   
   getAudioSpeedCorrectionMPV' <- liftIO $ readFile "jsons/getAudioSpeedCorrectionMPV.json"
   
   liftIO $ send (r^.mpvConnection) getAudioSpeedCorrectionMPV'
   (liftIO $ recv (r^.mpvConnection) 4096) 
      >>= (return . responseWrapper) 
      >>= (response L..= )
   
   r<- St.get
   when (r^.verbose>0) $ liftIO $ BC8.putStrLn $ fromMaybe "---" $ r^.response 
   
   L.assign mpvSpeed $ maybe Nothing parseForAudioSpeedCorrection (r^.response) 
   r<- St.get
   when (r^.verbose>2) $  liftIO $ print $ r^.mpvSpeed
   
   where
   responseWrapper "" = Nothing
   responseWrapper x  = Just $ BC8.pack x
   
   


   
setTime_posMPV :: StateT A IO ()
setTime_posMPV = do 
   r<- St.get
   when (r^.verbose>2) $  liftIO $ putStrLn comment
   
   jsonCommand <- liftIO $ readFile "jsons/setTime_posMPV.json"
   
   case ((r^.kodiPlayerPositionTime), (r^.mpvTime_pos)) of
      (Just (h,m,s,ms), Just mt) -> do
         let d = deltaT h m s ms mt
         
         when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
            ++ " case ((r^.kodiPlayerPositionTime), (r^.mpvTime_pos)) good"
         
         case (((abs d) > (r^.adjThreshold))||(r^.pause)) of
            False -> do
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
                  ++ " case ((abs d) > (r^.adjThreshold)) false"
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ ("delta t within range: " ++ (show d))
               
            True -> do 
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
                  ++ " case ((abs d) > (r^.adjThreshold)) true"
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ ("adjusting mpv time: " ++ (show d))
               
               let c = jsonCommand & key "command" . nth 2 . _Double .~ (realToFrac (mt + d))
               e<- liftIO $ send (r^.mpvConnection) c
               
               --L.assign mpvSpeed $ Just 1
               
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment ++ " case send: " ++ c
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ comment ++ " send response= " ++ show e
         
         
      err -> liftIO $ putStrLn $ ("cant set adjustment, time undefined: " ++ show err)
      
   where
   deltaT :: Int -> Int -> Int -> Int -> Float -> Float
   deltaT h m s ms mt = (fromIntegral ((60*60*h) + (60*m) + s)) + (read ( "0." ++ (show ms))) - mt
   comment = "setTime_posMPV"





setSpeedMPV :: StateT A IO ()
setSpeedMPV = do 
   r<- St.get
   when (r^.verbose>2) $  liftIO $ putStrLn comment
   
   jsonCommand <- liftIO $ readFile "jsons/setSpeedMPV.json"
   
   case ((r^.kodiPlayerPositionTime), (r^.mpvTime_pos), (r^.mpvSpeed)) of
      (Just (h,m,s,ms), Just mt, Just msc) -> do
         let d = deltaT h m s ms mt
         
         
         
         when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
            ++ " case ((r^.kodiPlayerPositionTime), (r^.mpvTime_pos)) good"
         
         --case True of
         case ((abs d) > 0.04) of
            False -> do
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
                  ++ " case ((abs d) > (r^.adjThreshold/3)) false"
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ ("delta t within range: " ++ (show d))
                 -- $ liftIO $ putStrLn $ ("adjusting mpv time: " ++ (show d))
                  
               L.assign desync d
               
            True -> do 
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment 
                  ++ " case ((abs d) > (r^.adjThreshold/3)) true"
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ ("adjusting mpv time: " ++ (show d))
               
               
               
               let c = jsonCommand & key "command" . nth 2 . _Double .~ 
                                                              --(realToFrac (msc + speedAdjustment d))
                                                  (realToFrac $ speedAdjustment' msc d $ r^.desync)
               e<- liftIO $ send (r^.mpvConnection) c
               
               L.assign desync d
               
               when (r^.verbose>3) $  liftIO $ putStrLn $ comment ++ " case send: " ++ c
               when ((r^.verbose)>2) 
                  $ liftIO $ putStrLn $ comment ++ " send response= " ++ show e
         
         
      err -> liftIO $ putStrLn $ ("cant set adjustment, time undefined: " ++ show err)
      
   where
   deltaT :: Int -> Int -> Int -> Int -> Float -> Float
   deltaT h m s ms mt = (fromIntegral ((60*60*h) + (60*m) + s)) + (read ( "0." ++ (show ms))) - mt
   
   speedAdjustment :: Float -> Float
   speedAdjustment d 
      |d>0 = 0.01
      |d<0 = -0.01
      |otherwise = 0
      
   speedAdjustment' :: Float -> Float -> Float -> Float
   speedAdjustment' s d d'
      |d>0 && d'>0 && d<d'= late
      |d<0 && d'<0 && d'<d= ahead
      |otherwise = s + (noMoreThen0_05 (d/5))
      where
         late  = s + (noMoreThen0_05 $ inXsec 5 d s   cs)   
         ahead = s + (noMoreThen0_05 $ inXsec 5 d s (-cs))  
         cs = (abs d') - (abs d) --converge speed
         
         noMoreThen0_05 d
            |abs d <= 0.05 = d
            |otherwise = (sign d) * 0.05
         
         sign d
            |d<0 = -1
            |d>0 = 1
            |otherwise = 0
         
         inXsec x d s cs = ((s - cs) + (cs / (x/(d/cs)))) - s
            
   comment = "setSpeedMPV"



requestBody b o = Streams.write (Just $ BB.fromByteString $ b) 
                                o


typicalKodiJSON_request = buildRequest1 $ do
      http POST "/jsonrpc"
      --setAccept "text/html"
      setAccept "application/json"



responseHandling c j n = receiveResponse c f
   where
   f p i = do
      xm <- Streams.read i
      case xm of
         Just x    -> j x
         Nothing   -> n








kodiSidePreparation :: StateT A IO ()
kodiSidePreparation = do
   r<- St.get

   c <- liftIO $ openConnection (r^.kodiSocketHostname) (r^.kodiSocketPort)
   kodiConnection L..= c
   
   kodiSidePreparationGetActivePlayers
   
               
               
               
               
kodiSidePreparationGetActivePlayers :: StateT A IO ()
kodiSidePreparationGetActivePlayers = do
   r<- St.get
   when (r^.verbose>3) $ liftIO $ putStrLn "kodiSidePreparationGetActivePlayers"
   getKodiActivePlayers' <- liftIO getKodiActivePlayers

   liftIO $ sendRequest (r^.kodiConnection) 
                        typicalKodiJSON_request 
                        (requestBody getKodiActivePlayers')
   
   (liftIO $ responseHandling (r^.kodiConnection) 
                              (return.Just ) 
                              (return $ Nothing)) >>= (response L..=)

   r<- St.get
   when (r^.verbose>0) $ liftIO $ BC8.putStrLn $ fromMaybe "---" $ r^.response 
   
   case (r^.response) of
         Nothing -> do
            (liftIO $ putStrLn "no response")
            playerid L..= Nothing
         Just r  -> playerid L..= (parseForKodiPlayerid r)
               


sendPauseMpv :: Bool -> StateT A IO ()
sendPauseMpv p = do
   r<- St.get
   when (r^.verbose>3) $ liftIO $ putStrLn comment
   
   jsonCommand <- liftIO $ readFile "jsons/setPauseMPV.json"
   
   let c = jsonCommand & key "command" . nth 2 . _Bool .~ p
   e<- liftIO $ send (r^.mpvConnection) c
   
   when (r^.verbose>3) $  liftIO $ putStrLn $ comment ++ " case send: " ++ c
   when ((r^.verbose)>2) 
      $ liftIO $ putStrLn $ comment ++ " send response= " ++ show e
   
   where 
      comment = "sendPauseMpv"
   
   



kodiGetPlayerPositionTime :: Int -> StateT A IO ()
kodiGetPlayerPositionTime p = do 
   r<- St.get
   when (r^.verbose>3) $ liftIO $ putStrLn "kodiGetPlayerPositionTime"
   getKodiPlayerPositionTime' <- liftIO $ getKodiPlayerPositionTime p
   --forever $ do
   liftIO $ sendRequest 
      (r^.kodiConnection)
      typicalKodiJSON_request 
      (requestBody $ getKodiPlayerPositionTime')
      
   (liftIO $ responseHandling (r^.kodiConnection) 
                              (return.Just ) 
                              (return $ Nothing)) >>= (response L..=)
                              
   r<- St.get
   let kppt = maybe Nothing 
                    parseForKodiPlayerPositionTime 
                    (r^.response) 
   
   
   case (isPaused kppt $ r^.kodiPlayerPositionTime) of
      True -> do 
         L.assign pause True
         when (r^.verbose>0) $ liftIO $ putStrLn "is paused"
         sendPauseMpv True
      False -> do 
         L.assign kodiPlayerPositionTime $ maybe Nothing 
                                                 parseForKodiPlayerPositionTime 
                                                 (r^.response) 
         r<- St.get
         when (r^.verbose>0) $ liftIO $ BC8.putStrLn $ fromMaybe "---" $ r^.response
         when (r^.verbose>0) $ liftIO $ print $ r^.kodiPlayerPositionTime
         
         case (r^.pause) of
            True -> do
               sendPauseMpv False
               L.assign pause False
            False -> return ()

   
   where
      isPaused (Just kppt) (Just kppt_prev) 
         |kppt == kppt_prev = True
         |otherwise = False
      isPaused _ _ = False


mpvSidePreparation :: StateT A IO ()
mpvSidePreparation = do
   r<- St.get
   soc <- liftIO $ socket AF_UNIX Stream 0
   liftIO $ connect soc $ SockAddrUnix $ r^.mpvSocket
   mpvConnection L..= soc
   



loop :: StateT A IO ()
loop = do 
   kodiSidePreparation
   mpvSidePreparation
   
   forever $ do
      r<- St.get
      case (r^.playerid) of
         Nothing -> do 
            liftIO $ putStrLn "no playerid found"
            kodiSidePreparationGetActivePlayers
         Just p -> do
            
            t1<- liftIO $ getPOSIXTime
            
            case ((r^.nextWakeupTime) < t1)of
               False -> do 
                  when (r^.verbose>4) $  liftIO 
                     $ putStrLn $ "wakeup at " 
                        ++ (show $ r^.nextWakeupTime) 
                        ++ " now " ++ (show t1)
                        ++ " remained " ++ (show (t1 - r^.nextWakeupTime))
                  
                  liftIO $ threadDelay 1000
               True -> do
                  when (r^.verbose>4) $  liftIO 
                     $ putStrLn $ "wakeup at " ++ (show $ r^.nextWakeupTime) ++ " now " ++ (show t1)
                  
                  L.assign nextWakeupTime (t1+1)
                  
                  
                  
                  kodiGetPlayerPositionTime p
                  
                  t2<- liftIO $ getPOSIXTime
                  getTime_posMPV
                  
                  t3<- liftIO $ getPOSIXTime
                  
                  let ping = t3-t1
                  
                  r<- St.get
                  
                  L.assign mpvTime_pos $ maybe 
                     (r^.mpvTime_pos) 
                     (\mt-> Just $ mt - realToFrac ping) 
                     (r^.mpvTime_pos)
                  
                  r<- St.get
                  when (r^.verbose>3) $  liftIO 
                     $ putStrLn $ "ping " ++ (show ping)
                  when (r^.verbose>3) $  liftIO 
                     $ putStrLn $ "mpv time with ping correction " ++ (show $ r^.mpvTime_pos)
                  
                  
                  setTime_posMPV
                  
                  getAudioSpeedCorrectionMPV
                  setSpeedMPV
                  
                  
                  {-r<- St.get
                  L.assign mpvSpeed -- $ maybe 
                     --(r^.mpvSpeed) 
                     $ Just 2.0
                  -}
                  
                  
                  
                  kodiSidePreparationGetActivePlayers

   r<- St.get
   liftIO $ closeConnection $ r^.kodiConnection
   liftIO $ sClose $ r^.mpvConnection
   
   


 
main :: IO ()
main = do
   a<- getArgs
   j<- readFile $ unwords a
   
   evalStateT loop $ A {
      _response = Nothing
     ,_playerid = Nothing
     ,_mpvSocket = fromMaybe "/tmp/mpvsocket" $ j ^? key "mpvSocket"  . _JSON
     ,_kodiSocketHostname = BC8.pack $ T.unpack $ fromMaybe "192.168.1.28" 
                                                           $ j ^? key "kodiSocketHostname" . _String
     ,_kodiSocketPort = fromMaybe 8080  $ j ^? key "kodiSocketPort" . _JSON
     ,_verbose = fromMaybe 0 $ j ^? key "verbose" . _JSON
     ,_mpvCommError = Nothing
     ,_adjThreshold = fromMaybe 1 $ j ^? key "adjThreshold" . _JSON
     ,_nextWakeupTime = 0
     ,_desync = 0
     ,_pause = False
     ,_kodiPlayerPositionTime = Nothing
    -- ,_kodiConnection :: Connection
    -- ,_mpvConnection  :: Connection
     }



   
   
   


